/////////////////////////////////////////////////////////////////////////////
// Purpose:     GSwxLua stub = wxLua + wxGhostscript
// Author:      Piotr Pianowski, Piotr Strzelczyk
// Created:     22.02.2008
// Copyright:   (c) 2008 BOP s.c.
// Licence:     GPL
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#ifdef __GNUG__
    #pragma implementation "gswxlua.h"
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/config.h" 

#include "ghostscript.h"

#include "wxlua/include/wxlua.h"

extern "C"
{
    #include "lualib.h"
}

#include "wxbind/include/wxbinddefs.h"
// WXLUA_DECLARE_BIND_STD=
WXLUA_DECLARE_BIND_WXLUA
WXLUA_DECLARE_BIND_WXBASE
WXLUA_DECLARE_BIND_WXCORE
//WXLUA_DECLARE_BIND_WXADV
//WXLUA_DECLARE_BIND_WXNET
#include "wxluaghost_bind.h"
      
  
#include "gswxlua.h"

//#include "ico/psvimg.xpm"  
//#include "ico/psvtxt.xpm"  

const wxChar* wxGsWindowDefaultName = wxT("GS window");
const wxChar* wxGhostscriptBanner = wxT("This is wxLuaGhostscscript.\n");

// Create a new application object: this macro will allow wxWindows to create
// the application object during program execution (it's better than using 
// a static object for many reasons) and also declares the accessor function
// wxGetApp() which will return the reference of the right type (i.e. gswxluaApp 
// and not wxApp)

// ============================================================================
// implementation
// ============================================================================

BEGIN_EVENT_TABLE(gswxConsole, wxFrame)
    EVT_CLOSE(gswxConsole::OnCloseWindow)
END_EVENT_TABLE()

gswxConsole::gswxConsole(wxWindow* parent, wxWindowID id, const wxString& title,
                           const wxPoint& pos, const wxSize& size,
                           long style, const wxString& name)
             :wxFrame(parent, id, title, pos, size, style, name)
{
   m_exit_on_error = false;
   m_textCtrl = new wxTextCtrl(this, wxID_ANY, wxEmptyString,
                                wxDefaultPosition, wxDefaultSize,
                                wxTE_MULTILINE | wxTE_READONLY | wxTE_RICH);
//   SetIcon(wxICON(psvtxt));
}

void gswxConsole::OnCloseWindow(wxCloseEvent &event)
{
    // Must NULL the console in the app so it won't be used.
    // Using EVT_DESTROY in the app causes a segfault if this is deleted
    // in wxApp::OnExit and through this is ugly, it works.
//  if (wxGetApp().m_gswxConsole == this)
//      wxGetApp().m_gswxConsole = NULL;
//
//   m_textCtrl->Destroy();
//   Destroy();
//   event.Skip();
     if (m_exit_on_error) Destroy();
     else Show(false);
}

void gswxConsole::DisplayText(const wxString& msg)
{
    m_textCtrl->AppendText(msg);
}

// Find the index into the argv that matches script name 
int gswxluaApp::FindFirstCmdParam(wxArrayString args)
{
   for (unsigned int arg_n = 1; arg_n < args.GetCount(); arg_n++) {
      if (! args[arg_n].StartsWith(wxT("-"))) return arg_n;
   }
   return -1; //oops?
}


// Push the program args into lua, see lua.c function
//    static int getargs (lua_State *L, char **argv, int n)
int gswxluaApp::PushArgs(wxArrayString args, int argc, int start_n)
{
    lua_State* L = m_wxlState.GetLuaState();

    int narg = argc - (start_n + 1);  // number of arguments to the script
//     extra arguments was pushed onto stack -- it looks obsolete???       
//        luaL_checkstack(L, narg + 3, "too many arguments to script");
//        for (int i = start_n+1; i < argc; i++)
//            lua_pushstring(L, wx2lua(argv[i]));

    lua_createtable(L, narg, argc - narg);

    for (int i = 0; i < argc; i++)
    {
       lua_pushstring(L, wx2lua(args[i]));
       lua_rawseti(L, -2, i - start_n);
    }
   
    lua_setglobal(L, "arg");
    return narg;
}


void gswxluaApp::DisplayMessage(const wxString &msg, bool is_error,
                                        const wxLuaState& wxlState)
{
//    if (m_want_console && ! (m_gswxConsole->IsShown()))
//    {
//        m_gswxConsole->Show(true);
//    }

    if (!is_error)
    {
        if (m_gswxConsole != NULL)
            m_gswxConsole->DisplayText(msg);
        else
        {
#ifdef __WXMSW__
//            wxMessageBox(msg, wxT("GSwxLua Print"));
#else
//           wxPrintf(msg);
           wxPrintf(wxT("%s"),msg.c_str());
#endif // __WXMSW__
        }
    }
    else
    {
        if (m_gswxConsole != NULL)
        {
            m_gswxConsole->DisplayText(msg);
            LeaveConsole();
        }
        else
        {
#ifdef __WXMSW__
            wxMessageBox(msg, wxT("GSwxLua Error"));
#else
           wxPrintf(wxT("%s"),msg.c_str());
#endif // __WXMSW__
        }
    }
}


void gswxluaApp::OnLua( wxLuaEvent &event )
{
    if (event.GetEventType() == wxEVT_LUA_ERROR)    
       DisplayMessage(wxString(wxGhostscriptBanner) + event.GetString(),true);
    else
       DisplayMessage(event.GetString(),false);
}


bool gswxluaApp::LeaveConsole() {
  if (m_gswxConsole) {
    m_gswxConsole->Show(true);
    m_gswxConsole->Iconize(false);
    m_gswxConsole->SetExitOnError(true);    
    return TRUE;
  } else return FALSE;
}

int CompileFile(const wxString &filename, wxString *errormsg)
{
    // create a new lua_State so we don't mess up our own
    lua_State *L = lua_open();
    luaL_openlibs(L); // load some useful libraries, loads all of them
    int status = luaL_loadfile(L,wx2lua(filename));
    *errormsg = wxString(lua2wx(lua_tostring(L, -1)));
    lua_close(L);
    return status;
}


// ----------------------------------------------------------------------------
// the application class
// ----------------------------------------------------------------------------

IMPLEMENT_APP(gswxluaApp);

gswxluaApp::gswxluaApp()
{
//  SetUseBestVisual(true); !!! makes bad scroll up
}

BEGIN_EVENT_TABLE(gswxluaApp, wxApp)
    EVT_LUA_PRINT       (wxID_ANY, gswxluaApp::OnLua)
    EVT_LUA_ERROR       (wxID_ANY, gswxluaApp::OnLua)
END_EVENT_TABLE()

// 'Main program' equivalent: the program execution "starts" here
bool gswxluaApp::OnInit()
{
   
#ifdef __WXGTK__
    // this call is very important since otherwise scripts using the decimal
    // point '.' could not work with those locales which use a different symbol
    // (e.g. the comma) for the decimal point...
    setlocale(LC_NUMERIC, "C");
#endif

    // Initialize the wxLua bindings we want to use.
    // See notes for WXLUA_DECLARE_BIND_STD above.
//    WXLUA_IMPLEMENT_BIND_STD;
    WXLUA_IMPLEMENT_BIND_WXLUA;
    WXLUA_IMPLEMENT_BIND_WXBASE;
    WXLUA_IMPLEMENT_BIND_WXCORE;
//    WXLUA_IMPLEMENT_BIND_WXADV;
//    WXLUA_IMPLEMENT_BIND_WXNET;
    wxLuaBinding_wxluaghost_init();

    m_want_console = false;
//  m_gswxConsole   = NULL;
//  console is always constructed (it can be hidden, but allways is defined)  
    m_gswxConsole = new gswxConsole(NULL, wxID_ANY);

    wxArrayString args = wxArrayString(argc,(const wxChar **)argv);
   
    wxString programName = args[0];
    wxFileName exeFileName=wxFileName(programName);
    exeFileName.Normalize();

    wxStandardPathsBase& stdpaths = wxStandardPaths::Get();
    if (! exeFileName.IsFileExecutable()) 
       exeFileName.Assign(stdpaths.GetExecutablePath());
     
    int arg_n,arg_gs_n;
    arg_gs_n=args.Index(wxT("--"));
    if (arg_gs_n==wxNOT_FOUND) arg_gs_n=args.GetCount()+1;
    

    arg_n=args.Index(wxT("--consola"));
    if (arg_n==wxNOT_FOUND) arg_n=args.Index(wxT("-c"));
    if (arg_n!=wxNOT_FOUND && arg_n<arg_gs_n) {
          m_want_console = true;
//          m_gswxConsole = new gswxConsole(NULL, wxID_ANY);
          m_gswxConsole->Show(true);
    }

    arg_n=args.Index(wxT("--lua"));
    if (arg_n==wxNOT_FOUND) arg_n=args.Index(wxT("-l"));
    if (arg_n==wxNOT_FOUND) arg_n=FindFirstCmdParam(args); else arg_n++;
    if (arg_n<0 || arg_n>=arg_gs_n) {
        DisplayMessage(wxString(wxGhostscriptBanner) + 
           wxT("To start a desired application use:\n  ") + programName + 
           wxT(" [-c|--console] [-l|--lua] <wxLua-application-name> [options]\n")
           + wxT("More help available when launched with a wxLua script\n"),true);
	return LeaveConsole();
    }
   
    wxString luaName=args[arg_n];
    wxFileName luaFileName=wxFileName(luaName);
//    luaFileName.Normalize(); 
//    luaFileName.IsFileReadable() -- doesn't work under WIN98 (mingw) -- to be tested
   
    if (! luaFileName.FileExists()) //.IsFileReadable()
       luaFileName.Assign(stdpaths.GetUserDataDir() + wxFileName::GetPathSeparator() 
		          + luaName);
   
    if (! luaFileName.FileExists()) //.IsFileReadable()
       luaFileName.Assign(stdpaths.GetDataDir() + wxFileName::GetPathSeparator() 
			  + luaName);
   
//  if (! luaFileName.IsFileReadable() && exeFileName.IsFileExecutable())
    if (! luaFileName.FileExists() && exeFileName.FileExists())
       luaFileName.Assign(exeFileName.GetPathWithSep() + luaName);
   
    if (luaFileName.FileExists()) {
       luaName=luaFileName.GetFullPath();
    } else {
       DisplayMessage(wxString(wxGhostscriptBanner) + 
          wxT("Lua error: Can't find Lua script: ") + luaName +
          wxT("\n"), true);
       return LeaveConsole();
    }
       
    m_wxlState = wxLuaState(this, wxID_ANY);
    if (!m_wxlState.Ok()) {
       DisplayMessage(wxString(wxGhostscriptBanner) + 
          wxT("Lua error: Can't initialize Lua state\n"), true);
       return LeaveConsole();
    }
 
    args[arg_n]=luaName;
    PushArgs(args, argc, arg_n); // PushArgs uses m_wxlState!!!
    
    wxString errmsg;
    int rc = CompileFile(luaName,&errmsg);
    if (rc!=0)  {
        DisplayMessage(wxString(wxGhostscriptBanner) +
           wxlua_LUA_ERR_msg(rc) + wxT("\n") + errmsg + wxT("\n") , true);
        return LeaveConsole();
    }

    rc = m_wxlState.RunFile(luaName);
    if (rc!=0 || GetTopWindow() == m_gswxConsole && m_gswxConsole->IsShown()) 
      return LeaveConsole();
    if (GetTopWindow() == NULL || GetTopWindow() == m_gswxConsole) return FALSE;
    // another Window was created so run it:
    return TRUE;
}

// Tidy up when program exits
int gswxluaApp::OnExit()
{
    // Delete application objects
    m_wxlState.CloseLuaState(true);
    m_wxlState.Destroy();
  
    return 0;
}

// ----------------------------------------------------------------------------
// Lua Ghostscript class
// ----------------------------------------------------------------------------


IMPLEMENT_DYNAMIC_CLASS(wxLuaGhostscript,wxGhostscript);

wxLuaGhostscript::wxLuaGhostscript() :wxGhostscript()
{m_wxlState=(wxLuaState*)NULL;}

wxLuaGhostscript::wxLuaGhostscript(const wxLuaState& wxlState)
 :wxGhostscript(), m_wxlState(wxlState) 
{}

wxLuaGhostscript::wxLuaGhostscript(const wxLuaState& wxlState, 
        wxGsDll *gsdll, const wxArrayString& argarray, ImageType imagetype) 
  :wxGhostscript(gsdll, argarray, imagetype), m_wxlState(wxlState)
{
//   Create(gsdll, argarray, imagetype);
}


void wxLuaGhostscript::OnStdout(wxString text)
{
    if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnStdout", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.lua_PushString(text);
        m_wxlState.LuaPCall(2, 0);
        m_wxlState.lua_SetTop(nOldTop);
    }
    else        
       wxGhostscript::OnStdout(text);
    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
}

void wxLuaGhostscript::OnStderr(wxString text)
{
    if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnStderr", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.lua_PushString(text);
        m_wxlState.LuaPCall(2, 0);
        m_wxlState.lua_SetTop(nOldTop);
    }
    else        
       wxGhostscript::OnStderr(text);
    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
}

wxString *wxLuaGhostscript::OnStdin()
{
    if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnStdin", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.LuaPCall(1, 1);
        GSin_wxstring = m_wxlState.GetwxStringType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        GSin_wxstring = *wxGhostscript::OnStdin();
   
    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return &GSin_wxstring;
}

int wxLuaGhostscript::OnPoll()
{
//    wxPrintf(wxT("!!poll"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnPoll", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.LuaPCall(1, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnPoll();

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplayOpen(wxGsImage& image)
{
//   wxPrintf(wxT("!!DisplayOpen"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayOpen", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayOpen(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplayPreClose(wxGsImage& image)
{
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayPreClose", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayPreClose(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}


int wxLuaGhostscript::OnDisplayClose(wxGsImage& image)
{
//   wxPrintf(wxT("!!DisplayClose"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayClose", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayClose(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplayPreSize(wxGsImage& image)
{
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayPreSize", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayPreSize(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}


int wxLuaGhostscript::OnDisplaySize(wxGsImage& image)
{
//   wxPrintf(wxT("!!DisplaySize"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplaySize", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplaySize(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplaySync(wxGsImage& image)
{
//   wxPrintf(wxT("!!DisplaySync"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplaySync", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.LuaPCall(2, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplaySync(image);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplayPage(wxGsImage& image, int copies, bool flush)
{
//   wxPrintf(wxT("!!DisplayPage"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayPage", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.lua_PushInteger(copies);
        m_wxlState.lua_PushBoolean(flush);
        m_wxlState.LuaPCall(4, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayPage(image,copies,flush);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}

int wxLuaGhostscript::OnDisplayUpdate(wxGsImage& image, wxRect region)
{
//   wxPrintf(wxT("!!DisplayUpdate"));
   int result;
   if (m_wxlState.Ok() && !m_wxlState.GetCallBaseClassFunction() &&
        m_wxlState.HasDerivedMethod(this, "OnDisplayUpdate", true))
    {
        int nOldTop = m_wxlState.lua_GetTop();
        m_wxlState.wxluaT_PushUserDataType(this, wxluatype_wxLuaGhostscript, true);
        m_wxlState.wxluaT_PushUserDataType(&image,wxluatype_wxGsImage,true);
        m_wxlState.wxluaT_PushUserDataType(&region,wxluatype_wxRect,true);
        m_wxlState.LuaPCall(3, 1);
        result = m_wxlState.GetIntegerType(-1);
        m_wxlState.lua_SetTop(nOldTop);
   }
    else
        result = wxGhostscript::OnDisplayUpdate(image,region);

    m_wxlState.SetCallBaseClassFunction(false); // clear flag always
    return result;
}


