/////////////////////////////////////////////////////////////////////////////
// Purpose:     GSwxLua stub = wxLua + wxGhostscript
// Author:      Piotr Pianowski, Piotr Strzelczyk
// Created:     22.02.2008
// Copyright:   (c) 2008 BOP s.c.
// Licence:     GPL
/////////////////////////////////////////////////////////////////////////////

#ifndef gswxluaApp_H
#define gswxluaApp_H

#ifdef __GNUG__
    #pragma interface "gswxlua.h"
#endif

#include "wx/cmdline.h"

#define VERSION "0.1"

  
class gswxConsole : public wxFrame
{
public:
    gswxConsole(wxWindow* parent, wxWindowID id,
                 const wxString& title = wxT("GSwxLua Console"),
                 const wxPoint& pos = wxDefaultPosition,
                 const wxSize& size = wxSize(800, 600),
                 long style = wxDEFAULT_FRAME_STYLE,
                 const wxString& name = wxT("GSwxLua Console"));

    // Display a message in the console
    void DisplayText(const wxString& msg);
    // An error has occurred, when this window is closed call wxExit to close the app.
    void SetExitOnError(bool is_error) { m_exit_on_error = m_exit_on_error || is_error; }

protected:
    void OnCloseWindow(wxCloseEvent& event);
    wxTextCtrl       *m_textCtrl;
    bool              m_exit_on_error;

private:
    DECLARE_EVENT_TABLE();
};



// Define a new application type, each program should derive a class from wxApp
class gswxluaApp : public wxApp
{
public:
    gswxluaApp();

    // override base class virtuals
    // ----------------------------

    // this one is called on application startup and is a good place for the app
    // initialization (doing it here and not in the ctor allows to have an error
    // return: if OnInit() returns false, the application terminates)
    virtual bool OnInit();
    virtual int OnExit();

    // Accessors
    void OnLua( wxLuaEvent &event );

    void DisplayMessage(const wxString &msg, bool is_error = false,
                        const wxLuaState& wxlState = wxNullLuaState);
    int FindFirstCmdParam(wxArrayString args);
    int PushArgs(wxArrayString args, int argc, int start_n);
    bool LeaveConsole();

    wxLuaState     m_wxlState;
    gswxConsole   *m_gswxConsole;
    bool           m_want_console;

private:
    DECLARE_EVENT_TABLE();
};

DECLARE_APP(gswxluaApp);
  
  
// Sub-class of Ghostscript class which relays callbacks to Lua
// -----------------------------------------------------------------------
class wxLuaGhostscript: public wxGhostscript
{
DECLARE_DYNAMIC_CLASS(wxLuaGhostscript);
public:
     wxLuaGhostscript();
     wxLuaGhostscript(const wxLuaState& wxlState);
     wxLuaGhostscript(const wxLuaState& wxlState, 
        wxGsDll *gsdll, const wxArrayString& argarray, ImageType imagetype = colour);
    ~wxLuaGhostscript() {}
    wxString *OnStdin();
    void OnStdout(wxString text);
    void OnStderr(wxString text);
    int OnPoll();
    int OnDisplayOpen(wxGsImage& image);
    int OnDisplayPreClose(wxGsImage& image);
    int OnDisplayClose(wxGsImage& image);
    int OnDisplayPreSize(wxGsImage& image);
    int OnDisplaySize(wxGsImage& image);
    int OnDisplaySync(wxGsImage& image);
    int OnDisplayPage(wxGsImage& image, int copies, bool flush);
    int OnDisplayUpdate(wxGsImage& image, wxRect region);
private:
    wxLuaState   m_wxlState;
    wxString     GSin_wxstring;
};

extern const wxChar* wxGsWindowDefaultName;


#endif // gswxluaApp_H

